import datetime
import random

from abc import ABC, abstractmethod

from lemmings.utils.parsing import duration


class RPS(ABC):
    def __init__(self, max_duration=None):
        self.duration = duration(max_duration)

    def update_duration(self, new_duration):
        self.duration = duration(new_duration)

    def check_duration(self, offset):
        return self.duration and duration(offset) > self.duration

    @abstractmethod
    def sleep_time(self, offset):
        pass

    def __str__(self):
        if self.duration:
            return f" in duration {self.duration}"
        return ""


class Constant(RPS):
    def __init__(self, min_wait, max_wait=None, max_duration=None):
        super().__init__(max_duration)
        self.min_wait = min_wait
        self.max_wait = max_wait if max_wait else min_wait

    def sleep_time(self, offset):
        if self.check_duration(offset):
            return None
        return self.min_wait + random.random() * (self.max_wait - self.min_wait)

    def __str__(self):
        return f"Rate[constant wait {self.min_wait}{super().__str__()}]"


class NoLimits(RPS):

    def sleep_time(self, offset):
        if self.check_duration(offset):
            return None
        return datetime.timedelta(0)

    def __str__(self):
        return f"Rate[no limits{super().__str__()}]"

class Linear(RPS):
    def __init__(self, rps, max_duration=None):
        super().__init__(max_duration)
        self.rps = rps

    def sleep_time(self, offset):
        if self.check_duration(offset):
            return None
        return datetime.timedelta(seconds=1 / self.rps)

    def __str__(self):
        return f"Rate[linear {self.rps} r/s{super().__str__()}]"

class Complex(RPS):
    def __init__(self, *rps):
        sum = duration(0)
        for r in rps:
            if not r.duration:
                raise Exception()
            sum += r.duration
        super().__init__(sum)
        self.list = rps

    def sleep_time(self, offset):
        for r in self.list:
            x = r.sleep_time(offset)
            if x:
                return x
            offset -= r.duration
        return None

    def __str__(self):
        return f"Rate[{', '.join(map(str, self.list))}]"



from lemmings.utils.colors import *
from lemmings.utils.const import *
from lemmings.utils.coroutine import *
from lemmings.utils.influx import *
from lemmings.utils.parsing import *
from lemmings.utils.prometheus import *
from lemmings.utils.timer import *
from lemmings.utils.utils import *


def _num(num):
    return f"{num:4d}"
    # return Color.underline(f"{num:4d}")
    # return Color.back(Color.cyan, f"{num:4d}")


import asyncio
import functools

from lemmings.utils.timer import Timer


class Task:
    DEFAULT = Timer("default")

    def __init__(self, func=None, weight=1, order=0, comment="", system=False):
        self.name = "noname"
        self.timer = Task.DEFAULT
        self.func = func
        self.weight = weight
        self.order = order
        self.comment = comment
        self.system = system

    def set_name(self, name):
        self.name = name

    def set_test_plan(self, test_plan):
        self.test_plan = test_plan
        if (self.timer.name == "default") and (self.name != "noname"):
            self.timer = Timer(f"{self.test_plan.__class__.__name__}_{self.name}".lower())

    # use object as decorator
    def __call__(self, func=None):
        if not callable(self.func):  # no arguments to decorator (only @Task)
            self.func = func
            return self
        if asyncio.iscoroutinefunction(self.func):
            @functools.wraps(self.func)
            async def wrapper_timer(*args, **kwargs):
                with self.timer:
                    with self:
                        return await self.func(*args, **kwargs)

            return wrapper_timer
        else:
            @functools.wraps(self.func)
            def wrapper_timer(*args, **kwargs):
                with self.timer:
                    with self:
                        return self.func(*args, **kwargs)()

            return wrapper_timer

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # if isinstance(value, BaseException):
        #     print(f"exception during {self.name}: {value}")
        # return isinstance(value, TypeError)
        return

    def __str__(self):
        return f"{self.name}[w={self.weight}]"



import asyncio
import datetime
import os
import sys
import time
import traceback
from asyncio.locks import BoundedSemaphore

from lemmings.task_set_run import TaskSetRun
from lemmings.utils import _num, duration2sec, exception_with_line
from lemmings.utils.colors import Color
from lemmings.utils.const import DONE, FAIL

program_start = time.time()


class Worker:
    total = 0

    def __init__(self, name, task_set, max_in_parallel=1000, ansi_colors=False, debug=False):
        Worker.total += 1
        self.debug = debug
        self.id = Worker.total
        self.name = name
        self.task_set = task_set
        self.ansi_colors = ansi_colors

        self.title = f"{self.name}[{self.id}]"
        self.max_in_parallel = max_in_parallel

        # will be init in execute (in child processes):
        self.stat = None
        self.shared_state = None
        self.semaphore = None

    def set_title(self, title):
        self.title = title

    def set_shared(self, shared_state):
        self.shared_state = shared_state

    def set_statistics(self, stat):
        self.stat = stat

    async def run_task(self, offset, task):
        await asyncio.sleep(duration2sec(offset))
        start = time.time()

        def print_stat(prefix, print=True, comment="", success=False, failed=False):
            succ = Color.mark(Color.green, _num(self.stat.executed))
            fail = Color.mark(Color.red, _num(self.stat.failed))
            if success:
                succ = Color.bold(succ)
            if failed:
                fail = Color.bold(fail)
            total = _num(self.stat.total)
            if task.comment:
                comment = task.comment + " " + comment
            if print:
                self.log(f"{prefix:5}> {succ} / {fail} / {total} in {time.time() - start:.2f} sec", comment)

        try:
            async with self.semaphore:
                print_stat("Run  ", self.debug)
                await task.__call__(task)(self.task_set, self.task_run)
                if not task.system:
                    self.stat.success()
                print_stat(DONE, success=True)
        except BaseException as e:
            if not task.system:
                self.stat.fail()
            print_stat(FAIL, comment=exception_with_line(e), failed=True)
            if self.debug:
                traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                traceback.print_exc()

    def execute(self):
        Color.enable(self.ansi_colors)
        loop = asyncio.get_event_loop()
        # offset = datetime.timedelta(0)
        scheduled_tasks = []

        def schedule(task, offset=datetime.timedelta(0)):
            future = asyncio.ensure_future(self.run_task(offset, task))
            scheduled_tasks.append(future)
            if not task.system:
                self.stat.schedule()

        self.task_run = TaskSetRun(self, self.task_set)
        self.semaphore = self._semaphore()

        # setup
        for x in self.task_run.schedule(False):
            offset, task = x
            schedule(task, offset)
        self.log(f"Queue: {_num(self.stat.total)} in worker / {_num(self.stat.shared_total)} total")
        loop.run_until_complete(asyncio.wait(scheduled_tasks))

        # teardown
        scheduled_tasks = []
        schedule(self.task_run.teardown_task())
        loop.run_until_complete(asyncio.wait(scheduled_tasks))

        self.dump_results()
        return os.getpid()

    def _semaphore(self):
        p = self.max_in_parallel
        if self.task_run.is_ordered:  # no concurrent run in ordered execution
            p = 1
        return BoundedSemaphore(p)

    def dump_results(self):
        with self.stat.lock:
            self.log(f"worker results: {_num(self.stat.executed)} {DONE},"
                     f" {_num(self.stat.failed)} {FAIL} "
                     f"[{_num(self.stat.total)} total]")
            self.log(f"total  results: {_num(self.stat.shared_executed)} {DONE},"
                     f" {_num(self.stat.shared_failed)} {FAIL} "
                     f"[{_num(self.stat.shared_total)} total]")

    def log(self, msg, comment=""):
        global program_start
        if comment:
            comment = " | " + Color.cursive(comment)
        Color.print(self.id, f'{self}', f" {msg} {comment}", prefix=f"| {(time.time() - program_start):7.3f} ")

    def __str__(self):
        return self.title



from multiprocessing import Lock

from lemmings.shared import SharedValue


class Statistics:
    def __init__(self):
        self.lock = Lock()
        self.processed = 0
        self.executed = 0
        self.failed = 0
        self.total = 0
        self.shared_executed = SharedValue(0)
        self.shared_failed = SharedValue(0)
        self.shared_total = SharedValue(0)  # .with_name("total") # debug

    def schedule(self):
        self.total += 1
        self.shared_total += 1

    def success(self):
        self.processed += 1
        self.executed += 1
        self.shared_executed += 1

    def fail(self):
        self.processed += 1
        self.failed += 1
        self.shared_failed += 1
